/**************************************************************************/
/*                                                                        */
/*                                 OCaml                                  */
/*                                                                        */
/*             Xavier Leroy, projet Cristal, INRIA Rocquencourt           */
/*                                                                        */
/*   Copyright 1996 Institut National de Recherche en Informatique et     */
/*     en Automatique.                                                    */
/*                                                                        */
/*   All rights reserved.  This file is distributed under the terms of    */
/*   the GNU Lesser General Public License version 2.1, with the          */
/*   special exception on linking described in the file LICENSE.          */
/*                                                                        */
/**************************************************************************/

#include <caml/fail.h>
#include <caml/mlvalues.h>
#include "caml/unixsupport.h"

#ifdef HAS_SOCKETS

#include "caml/socketaddr.h"

CAMLprim value caml_unix_getsockname(value sock)
{
  int retcode;
  struct sockaddr_storage addr;
  socklen_t addr_len;

  addr_len = sizeof(addr);
  retcode = getsockname(Int_val(sock), (struct sockaddr *) &addr, &addr_len);
  if (retcode == -1) caml_uerror("getsockname", Nothing);
  return caml_unix_alloc_sockaddr((struct sockaddr *) &addr, addr_len, -1);
}

#else

CAMLprim value caml_unix_getsockname(value sock)
{ caml_invalid_argument("getsockname not implemented"); }

#endif
